import sys
import abc
import functools
import pathlib
import string
from typing import Union, Optional

from . import gift
from . import colors
from . import parsing
from . import transformer


def user_settings_to_class_init(settings: dict, name: Optional[str] = None) -> str:
	"""
	Turns a user settings dictionary into one that can be passed to a question's  `__init__`.

	Parameters
	----------
	settings : dict
		User settings.
	name : str, optional
		The name of the question.

	Returns
	-------
	out: str
		The name of the class to be instantiated.

	"""

	if name:
		settings['name'] = name

	return settings.pop('class')


class HtmlQuestion(metaclass=abc.ABCMeta):
	"""
	Abstract class implementing an html-based question.
	"""

	# in order to process LaTeX commands *in text* (ignoring occurrences in formulas)
	latex_commands_within_text_processor = transformer.LatexCommandsWithinText()

	def __init__(
			self, name: str, statement: str, images_settings: Optional[dict] = None, feedback: Optional[str] = None,
			time: Optional[int] = None, pre_transforms: list = [], post_transforms: list = []):
		"""
		Initializer.

		Parameters
		----------
		name : str
			Name of the question.
		statement : str
			Statement of the question.
		images_settings : dict, optional
			width and height of *all* the images in the question.
		feedback : str, optional
			Feedback for the question.
		time: int, optional
			The number of minutes deemed necessary to answer the question.
		pre_transforms: list
			Processors to apply before everything else.
		post_transforms: list
			Processors to apply in the end.
		"""

		self.name = name
		self.statement = statement.rstrip()
		self.feedback = feedback
		self.time = time
		self.pre_transforms = pre_transforms
		self.post_transforms = post_transforms

		# transformations that depend on particular settings of this class
		self.custom_transforms = [transformer.URLs(images_settings)]

	def process_text(self, text: str) -> str:
		"""
		Functions in `self.processing_functions` are applied on the given input.

		Parameters
		----------
		text : str
			Input text.

		Returns
		-------
		out: str
			Text processed with the previously defined functions.

		"""

		for function in (self.pre_transforms + self.custom_transforms + self.post_transforms):

			try:

				text = function(text)

			except gift.NotCompliantLatexFormula as e:

				print(
					f'\n{colors.error}cannot compile latex formula\n {colors.extra_info}{e.formula}{colors.reset} in '
					f'{colors.info}{self.name}')

				sys.exit(1)

		return text

	@property
	def gift(self):
		"""
		Builds the question in the GIFT format.

		Returns
		-------
		out: str
			GIFT-ready text.

		"""

		# so that the original statement is not modified
		statement = self.statement

		# feedback part of the answer if any
		feedback = ("\n\t" + gift.from_feedback(self.process_text(self.feedback.rstrip()))) if self.feedback else ""

		# the full answer
		answer = f'{{\n{self.answer + feedback}\n}}'

		# if a `time` estimate was passed...
		if self.time:

			# ...it is appended at the end of the statement
			statement += '\n\n\n' + r'<i>Estimated time\: ' + str(self.time) + r' minutes</i>' + '\n'

		return gift.from_question_name(self.name) + gift.html + self.process_text(statement) + answer

	def __repr__(self):

		return f'Name: {self.name}\nStatement: {self.statement}'

	@property
	@abc.abstractmethod
	def answer(self):
		"""
		Abstract method yielding the GIFT-ready answer to the question.

		Returns
		-------
		out: str
			GIFT-ready text with the answer.

		"""

		pass


class Numerical(HtmlQuestion):
	"""
	Class implementing a numerical question.
	"""

	def __init__(
			self, name: str, statement: str, solution: dict, images_settings: Optional[dict] = None,
			feedback: Optional[str] = None, time: Optional[int] = None, pre_transforms: list = [],
			post_transforms: list = []):
		"""
		Initializer.

		Parameters
		----------
		solution : dict
			Value and, optionally, tolerated error of the solution.
		"""

		super().__init__(name, statement, images_settings, feedback, time, pre_transforms, post_transforms)

		assert ('value' in solution), '"value" missing in "solution"'

		self.solution_value = solution['value']

		if 'error' in solution:

			# try to match a percentage
			m = parsing.re_percentage.match(str(solution['error']))

			# if so...
			if m:

				self.solution_error = str(solution['value'] * float(m.group(1)) / 100.)

			else:

				self.solution_error = str(solution['error'])

		# an error was NOT provided
		else:

			self.solution_error = None

	def __repr__(self):

		error = f' (error: {self.solution_error[1:]})' if self.solution_error != '' else ''

		return super().__repr__() + '\n' + f'Solution: {self.solution_value}' + error

	@property
	def answer(self):

		return gift.from_numerical_solution(self.solution_value, self.solution_error)


class MultipleChoice(HtmlQuestion):
	"""
	Class implementing a multiple-choice question.
	"""

	template_wrong_answers = string.Template(r"**<font color='$color'>$text</font>**")

	def __init__(
			self, name: str, statement: str, answers: dict, images_settings: Optional[dict] = None,
			feedback: Optional[str] = None, time: Optional[int] = None, pre_transforms: list = [],
			post_transforms: list = []):
		"""
		Initializer.

		Parameters
		----------
		answers : dict
			Right answer and a list with the wrong ones.
		"""

		super().__init__(name, statement, images_settings, feedback, time, pre_transforms, post_transforms)

		self.answers = answers

	@property
	def answer(self):

		processed_answers = []

		# the maximum grade allowed by partially correct answers
		max_grade = 0.

		for a in self.answers['wrong']:

			# if it is a list
			if isinstance(a, list):

				processed_answers.append(f'~%{a[1]}%{self.process_text(a[0])}')

				# if the grade is positive...
				if a[1] > 0:

					# ...it is accounted when computing the maximum
					max_grade += float(a[1])

			# if it is a scalar
			else:

				processed_answers.append(f'~{self.process_text(a)}')

		# if a "perfect" (100% credit) was provided...
		if 'perfect' in self.answers:

			# ...it is added in a special way
			processed_answers.insert(0, '=' + self.process_text(self.answers['perfect']))

		else:

			# if it's not possible to get full credit...
			if max_grade < 100.:
				print(
					f"{colors.extra_info}question{colors.reset} \"{self.name}\" {colors.extra_info} won't allow "
					f"full credit {colors.reset}{max_grade}")

		return '\t' + '\n\t'.join(processed_answers)



from typing import Union, Optional

from . import latex


class NotCompliantLatexFormula(Exception):

	def __init__(self, formula: str) -> None:

		self.formula = formula

	def __str__(self) -> str:

		return self.formula


html = '[html]'


def from_question_name(name: str) -> str:
	"""
	Generates GIFT text for a question name.

	Parameters
	----------
	name : str
		Name of a question.

	Returns
	-------
	out: str
		GIFT-ready question name.

	"""

	return f'::{name}::'


def from_image_url(url: str, width: int, height: int) -> str:
	"""
	Generates GIFT text for a URL.

	Parameters
	----------
	url : str
		Input URL.
	width : int
		Desired width of the image.
	height : int
		Desired height of the image.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	# either both are set or none
	assert not ((width is not None) ^ (height is not None)), 'both "width" and "height" should be set or none of them'

	# `width` is checked, but `height` could be as well used (at this point either both are set or none)
	if width is not None:

		# notice the space at the beginning
		width_height = f' width="{width}" height="{height}"'

	else:

		width_height = ''

	res = f'<img src="{url}" alt="" role="presentation" class="atto_image_button_text-bottom"{width_height}>'

	for to_be_escaped in [':', '~', '=']:

		res = res.replace(to_be_escaped, '\\' + to_be_escaped)

	return res


def from_category(name: str, within_the_course: bool = True) -> str:
	"""
	Generates GIFT text for a category.

	Parameters
	----------
	name : str
		Category name.
	within_the_course: bool
		If True, the category will belong to the course in which the questions are imported.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	return '$CATEGORY: ' + (r'$course$/' if within_the_course else '') + name + '\n\n'


def from_feedback(text: str) -> str:
	"""
	Generates GIFT-ready text from a feedback string.

	Parameters
	----------
	text : str
		Input text.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	return '#'*4 + text


def from_latex_formula(text: str) -> str:
	"""
	Adapts a LaTeX formula to GIFT.

	Parameters
	----------
	text: str
		Naked (no $'s) LaTeX formula.

	Returns
	-------
	res: str
		GIFT-ready text.

	"""

	res = text

	for to_be_escaped in ['\\', '{', '}', '=']:

		res = res.replace(to_be_escaped, '\\' + to_be_escaped)

	res = res.replace('&', '&amp;')

	return r'\\(' + res + r'\\)'


def process_new_lines(text: str) -> str:
	"""
	Adapts every new line to GIFT.

	Parameters
	----------
	text : str
		Input text.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	# new lines are replaced everywhere *except* inside latex formulas
	return latex.replace_and_replace_only_in_formulas('\n', r'<br>', r'<br>', ' ', text)


def from_numerical_solution(solution: [float, int], error: Optional[Union[float, int, str]] = None) -> str:
	"""
	Generates GIFT-ready text from the solution and, optionally, error of a numerical question.

	Parameters
	----------
	solution : int or float
		Value of the solution.
	error : int or float or str, optional
		Value of the error.

	Returns
	-------
	out: str
		GIFT-ready text.

	"""

	error = ':' + str(error) if error else ''

	return f'#\t=%100%{solution}{error}#'



import re
import pathlib
import functools
from typing import Union, Callable, Optional

from . import image
from . import parsing
from . import remote
from . import gift
from . import latex


def process_paths(
		text: str, pattern: str, process_match: Callable[[str], None], replacement: Union[str, Callable[..., str]]):
	"""
	It searches in a text for strings corresponding to files (maybe including a path), replaces them by another
	string according to some function and, additionally, processes each file according to another function.

	Parameters
	----------
	text : str
		Input text.
	pattern : str
		Regular expression including a capturing group that yields the file.
	process_match : Callable[[str], None]
		Function that will *process* each file.
	replacement : str or Callable[[re.Match], str]
		Regular expression making use of the capturing group or function processing the match delivered by `pattern`

	Returns
	-------
	out: str
		Output text with the replacements performed, after having processed all the files.

	"""

	# all the matching files in the given text
	files = re.findall(pattern, text)

	# breakpoint()

	# every one of them...
	for file in files:
		# ...is processed
		process_match(file)

	# replacement of matches
	return re.sub(pattern, replacement, text)


class Transformer:

	def __init__(self) -> None:

		# subclasses are expected to set this up
		self.function: Optional[Callable] = None

	def __call__(self, text: str):

		assert self.function is not None, 'method "function" was not defined'

		return self.function(text)


class TexToSvg(Transformer):
	"""
	Processor to convert TeX files into svg files.
	"""

	def __init__(self, history: dict) -> None:

		super().__init__()

		self.history = history

		def process_match(f):

			# if this file has not been already compiled-converted...
			if f not in self.history['already compiled']:

				# ...it is...
				image.pdf_to_svg(image.tex_to_pdf(f))

				# ...and a note is made of it
				self.history['already compiled'].add(f)

		# (the "\1" in `replacement` refers to matches in `pattern`)
		self.function = functools.partial(
			process_paths, pattern=parsing.tex_file_name, process_match=process_match, replacement=r'\1.svg')


class SvgToHttp(Transformer):
	"""
	Processor to transfer svg files to a remote location.
	"""

	def __init__(
			self, history: dict, connection: remote.Connection, public_filesystem_root: str,
			pictures_base_directory: str, public_url: str):

		super().__init__()

		self.history = history

		# assembled remote path
		remote_subdirectory = pathlib.Path(public_filesystem_root).joinpath(pictures_base_directory)

		# TODO: when minimum Python version is forwarded to 3.8, `re.Match` should be the type hinting for `m`
		def replacement_function(m) -> str:

			file = pathlib.Path(m.group(0))

			return public_url + pictures_base_directory + '/' + file.as_posix()

		def process_match(f):

			# if this file has not been already transferred...
			if f not in self.history['already transferred']:

				# ...it is...
				connection.copy(f, remote_directory=remote_subdirectory / pathlib.Path(f).parent)

				# ...and a note is made of the fact
				self.history['already transferred'].add(f)

		self.function = functools.partial(
			process_paths, pattern=parsing.url_less_svg_file, process_match=process_match,
			replacement=replacement_function)


class SvgToInline(Transformer):
	"""
	Processor to directly include svg files into a question.
	"""

	def __init__(self):

		super().__init__()

		# TODO: when minimum Python version is forwarded to 3.8, `re.Match` should be the type hinting for `m`
		def replacement_function(m) -> str:

			file = pathlib.Path(m.group(1))

			return image.svg_to_html(file)

		self.function = functools.partial(
			process_paths, pattern=parsing.svg_file, process_match=lambda x: None, replacement=replacement_function)


class URLs(Transformer):
	"""
	Processor to arrange URLs into a GIFT-appropriate format.
	"""

	url = f'http({parsing.regex_url_valid_character}+)(?!{parsing.regex_url_valid_character})'

	def __init__(self, images_settings: Optional[dict] = None):

		super().__init__()

		if images_settings is None:

			self.images_width, self.images_height = None, None

		else:

			assert ('width' in images_settings) and ('height' in images_settings), \
				'"width" and/or "height" missing in "image_settings"'

			self.images_width, self.images_height = images_settings['width'], images_settings['height']

		self.function = lambda text: re.sub(self.url, self.replacement, text)

	# TODO: when minimum Python version is forwarded to 3.8, `re.Match` should be the type hinting for "m"
	def replacement(self, m) -> str:

		return '<p>' + gift.from_image_url(m.group(0), width=self.images_width, height=self.images_height) + '<br></p>'


class LatexCommandsWithinText(Transformer):

	# in every list, the first pair of elements are the search pattern and replacement to be applied *globally* whereas
	# the second one will only be applied inside LaTeX formulas
	# NOTE: "\" is duplicated because it is assumed it has been escaped previously by `gift.process_latex`
	patterns = [
		# \textbf
		[r'\\textbf{([^}]+)}', r'<b>\1</b>', r'<b>([^<]*)</b>', r'\\textbf{\1}'],
		# \textit
		[r'\\textit{([^}]+)}', r'<i>\1</i>', r'<i>([^<]*)</i>', r'\\textit{\1}']
	]

	def __init__(self) -> None:

		super().__init__()

		def f(text: str) -> str:

			res = text

			for pat in self.patterns:

				res = latex.replace_and_replace_only_in_formulas(*pat, res)

			return res

		self.function = f


class LatexFormulas(Transformer):

	latex_formula = r'\$([^\$]*)\$'

	def __init__(self, latex_auxiliary_file: Union[str, pathlib.Path], check_compliance: bool) -> None:

		super().__init__()

		self.latex_auxiliary_file = latex_auxiliary_file
		self.check_compliance = check_compliance

		self.function = lambda text: re.sub(self.latex_formula, self.replacement, text)

	# TODO: when minimum Python version is forwarded to 3.8, `re.Match` should be the type hinting for "m"
	def replacement(self, m) -> str:

		latex_source = m.group(1)

		if self.check_compliance:

			if not latex.formula_can_be_compiled(latex_source, auxiliary_file=self.latex_auxiliary_file):
				raise gift.NotCompliantLatexFormula(latex_source)

		return gift.from_latex_formula(latex_source)



import re

# regular expression to extract a percentage
re_percentage = re.compile(r'(\d*\.\d+|\d+)\s*%')

regex_filename_valid_character = r'[\\/\w_\-]'
regex_url_valid_character = regex_filename_valid_character[:-1] + r'\:\.~]'

# ---------- tex files

tex_file_name = r'(\S+)\.tex'

# ---------- svg files

# svg_file = r'(\S+\.svg)'
svg_file = f'({regex_filename_valid_character}+\.svg)'

url_less_svg_file = (
		f'(?<!{regex_filename_valid_character})(?!http)'
		f'({regex_filename_valid_character}+\.svg)(?!{regex_filename_valid_character})')

re_svg_id = re.compile(r'id="([\w-]+)"')

# ---------- latex

latex_formula_with_no_capturing = r'\$[^\$]*\$'
